#include <Python.h>
#include "tao.h"

#define ERROR(x) { PyErr_SetString(PyExc_RuntimeError,x); printf("*** %s:%d: %s\n",__FILE__,__LINE__,x); error=1; return 0; }

#define SET_C_FLOAT_ARRAY(n,x,px) { \
  for (int i=0; i < (n); ++i) { \
    PyObject *pxi = PyList_GET_ITEM(px,i); \
    if (!PyFloat_Check(pxi)) ERROR("list must contain only floats"); \
    (x)[i] = PyFloat_AS_DOUBLE(pxi); \
  } \
}

#define SET_PY_FLOAT_ARRAY(n,px,x) { \
  for (int i=0; i < (n); ++i) { \
    PyObject *pxi = PyFloat_FromDouble((x)[i]); \
    if (!PyFloat_Check(pxi)) ERROR("list must be only floats"); \
    PyList_SetItem(px,i,pxi); \
  } \
}

static int error = 0;
static TAO_SOLVER tao;
static TAO_APPLICATION taoapp;

typedef struct {
  PyObject *px, *ogfun, *pxub, *pxlb;
  char *solver;
  double fatol, frtol, catol, crtol;
  int verbosity, maxits;
  int n;
} argType;

int Monitor(TAO_SOLVER tao,void *ptr)
{
  argType *arg = (argType*)ptr;
  int iterate;
  double f, gnorm, cnorm, xdiff;
  TaoTerminateReason reason;
  Vec X;
  PetscScalar *x;

  TaoGetSolutionStatus(tao,&iterate,&f,&gnorm,&cnorm,&xdiff,&reason);
  if (arg->verbosity > 0) {
    printf("Round: %4d  Obj: %12.5e  GradNorm: %11.5e\n",iterate,f,gnorm);
    if (arg->verbosity > 1) {
      TaoAppGetSolutionVec(taoapp,&X);
      VecGetArray(X,&x);
      printf("Solution: [");
      for (int i=0; i < arg->n; ++i)
	printf(" %.2e",x[i]);
      printf(" ]\n");
      VecRestoreArray(X,&x);
    }
  }
  return 0;
}


int
ObjectiveGradient(TAO_APPLICATION taoapp,Vec X,double *f, Vec G,void *ptr)
{
  argType *arg = (argType *) ptr;  
  double ff=0.0;
  PetscScalar *x,*g;

  if (error) return 0;

  /* Get pointers to vector data */
  VecGetArray(X,&x);
  VecGetArray(G,&g);

  PyObject *px = PyList_New(arg->n);
  SET_PY_FLOAT_ARRAY(arg->n,px,x);
  PyObject *ret = PyObject_CallFunction(arg->ogfun,"O",px);
  if (PyErr_Occurred()) {
    error=1;
    for (int i=0; i < arg->n; ++i) { x[i] = 0.0; g[i] = 0.0; }
    VecRestoreArray(X,&x);
    VecRestoreArray(G,&g);
    //TaoSetTerminationReason(tao,TAO_DIVERGED_USER);
    return 0;
  }
  Py_DECREF(px);
  if (!ret) ERROR("ogfun: run-time error or bad argument spec");
  PyObject *pg;
  if (!PyArg_ParseTuple(ret,"dO",&ff,&pg))
    ERROR("ogfun returned a bad value: should be (fobj,g)");
  Py_INCREF(pg);
  Py_DECREF(ret);
  if (!PyList_Check(pg)) ERROR("ogfun returned bad gradient (not a list)");
  if (PyList_GET_SIZE(pg) != arg->n) ERROR("gradient list is wrong size");
  SET_C_FLOAT_ARRAY(arg->n,g,pg);
  Py_DECREF(pg);

  /* Restore vectors */
  VecRestoreArray(X,&x);
  VecRestoreArray(G,&g);
  *f=ff;
  return 0;
}

int
PrintBounds()
{
  TaoVecDoubleArray *txub, *txlb;
  TaoScalar *xub, *xlb;
  int nub, nlb;

  TaoGetVariableBounds(tao,(TaoVec**)&txlb,(TaoVec**)&txub);
  txub->GetArray(&xub,&nub);
  txlb->GetArray(&xlb,&nlb);
  assert(nub==nlb);
  for (int i=0; i < nub; ++i)
    printf("%f <= x[%d] <= %f\n",xlb[i],i,xub[i]);
  return 0;
}


int
SetBounds(TAO_APPLICATION taoapp, Vec XLB, Vec XUB, void *ptr)
{
  argType *arg = (argType *)ptr;
  PetscScalar *xub, *xlb;

  VecGetArray(XUB,&xub);
  VecGetArray(XLB,&xlb);

  SET_C_FLOAT_ARRAY(arg->n,xub,arg->pxub);
  SET_C_FLOAT_ARRAY(arg->n,xlb,arg->pxlb);

  VecRestoreArray(XUB,&xub);
  VecRestoreArray(XLB,&xlb);
  return 0;
}

int
ProcessArgs(PyObject *tuple, PyObject *dict, argType *arg)
{
  /* PARAMETER VALUES */
  arg->px = Py_None;
  arg->ogfun = Py_None;
  arg->pxub = Py_None;
  arg->pxlb = Py_None;
  arg->solver = "tao_lmvm";
  arg->fatol = 1e-4;
  arg->frtol = 1e-4;
  arg->catol = 0.0;
  arg->crtol = 0.0;
  arg->verbosity = 0;
  arg->maxits = 1000;

  char* argnames[] = {"x","ogfun","xub","xlb","solver","fatol","frtol",
		      "catol","crtol","verbosity","maxits",NULL};
  if (!PyArg_ParseTupleAndKeywords(tuple,dict,"OO|OOsddddii",argnames,
				   &arg->px,&arg->ogfun,&arg->pxub,&arg->pxlb,
				   &arg->solver,&arg->fatol,&arg->frtol,
				   &arg->catol,&arg->crtol,
				   &arg->verbosity,&arg->maxits))
    ERROR("Bad argument list to pytao.optimize()");
  Py_INCREF(arg->px);
  Py_INCREF(arg->ogfun);
  Py_XINCREF(arg->pxub);
  Py_XINCREF(arg->pxlb);

  if (!PyList_Check(arg->px)) ERROR("x is not a list");
  arg->n = PyList_GET_SIZE(arg->px);
  if (!PyCallable_Check(arg->ogfun)) ERROR("ogfun is not callable");
  if ((arg->pxub != Py_None) && (arg->pxlb != Py_None)) {
    if (arg->pxlb == Py_None || arg->pxub == Py_None)
      ERROR("xlb and xub must be passed together");
    if (!PySequence_Check(arg->pxub)) ERROR("xub is not a sequence");
    if (!PySequence_Check(arg->pxlb)) ERROR("xlb is not a sequence");
    if (PySequence_Length(arg->pxub) != arg->n) ERROR("len(xub) != len(x)");
    if (PySequence_Length(arg->pxlb) != arg->n) ERROR("len(xlb) != len(x)");
  }
  return 0;
}

static PyObject*
GetArgs(PyObject *self, PyObject *tuple, PyObject *dict)
{
  argType arg;
  ProcessArgs(tuple,dict,&arg);
  return Py_BuildValue("OOOOsddddii",arg.px,arg.ogfun,arg.pxub,arg.pxlb,
		       arg.solver,arg.fatol,arg.frtol,
		       arg.catol,arg.crtol,
		       arg.verbosity,arg.maxits);
}

void
PrintTerminationReason(TaoTerminateReason reason)
{
  char *s = "unknown (?)";
  if (reason == TAO_CONVERGED_ATOL) s = "res <= atol (success)";
  else if (reason == TAO_CONVERGED_RTOL) s = "res/res0 <= rtol (success)";
  else if (reason == TAO_CONVERGED_TRTOL) s = "xdiff <= trtol (success)";
  else if (reason == TAO_CONVERGED_MINF) s = "f <= fmin (success)";
  else if (reason == TAO_CONVERGED_USER) s = "CONVERGED_USER (success)";
  else if (reason == TAO_DIVERGED_MAXITS) s = "its > maxits (failure)";
  else if (reason == TAO_DIVERGED_NAN) s = "NaN/Numerical issues (failure)";
  else if (reason == TAO_DIVERGED_MAXFCN) s = "nfunc > maxnfuncts (failure)";
  else if (reason == TAO_DIVERGED_LS_FAILURE) s = "line search failure";
  else if (reason == TAO_DIVERGED_TR_REDUCTION)
    s = "DIVERGED_TR_REDUCTION (failure)";
  else if (reason == TAO_DIVERGED_USER) s = "DIVERGED_USER (failure)";
  else if (reason == TAO_CONTINUE_ITERATING) s = "CONTINUE_ITERATING (?)";
  printf("Termination reason: %s\n",s);
}

static PyObject*
Optimize(PyObject *self, PyObject *tuple, PyObject *dict)
{
  argType arg;
  Vec        X;                     /* solution vector */
  TaoTerminateReason reason;
  PetscScalar *x;

  ProcessArgs(tuple,dict,&arg);
  VecCreateSeq(PETSC_COMM_SELF,arg.n,&X);
  TaoCreate(PETSC_COMM_SELF,arg.solver,&tao);
  TaoSetMonitor(tao,Monitor,&arg);
  TaoSetMaximumIterates(tao,arg.maxits);
  TaoSetTolerances(tao,arg.fatol,arg.frtol,arg.catol,arg.crtol);
  TaoApplicationCreate(PETSC_COMM_SELF,&taoapp);
  if (arg.pxub) TaoAppSetVariableBoundsRoutine(taoapp,SetBounds,&arg);
  VecGetArray(X,&x);
  SET_C_FLOAT_ARRAY(arg.n,x,arg.px);
  VecRestoreArray(X,&x);
  Py_DECREF(arg.px);
  arg.px = NULL;
  TaoAppSetInitialSolutionVec(taoapp,X);
  TaoAppSetObjectiveAndGradientRoutine(taoapp,ObjectiveGradient,(void *)&arg);
  TaoSetOptions(taoapp,tao);
  TaoSolveApplication(taoapp,tao);
  TaoGetTerminationReason(tao,&reason);
  if (error) printf("***** Internal error\n");
  else if (reason < 0) PrintTerminationReason(reason);
  else if (arg.verbosity > 0) PrintTerminationReason(reason);
  VecGetArray(X,&x);
  PyObject *px = PyList_New(arg.n);
  SET_PY_FLOAT_ARRAY(arg.n,px,x);
  VecRestoreArray(X,&x);
  TaoAppDestroy(taoapp);
  TaoDestroy(tao);
  VecDestroy(X);
  Py_DECREF(arg.ogfun);
  Py_XDECREF(arg.pxub);
  Py_XDECREF(arg.pxlb);
  if (error)
    return 0;
  else
    return px;
}

static char optimize_docs[] =
"Python interface to TAO optimization package.\n\n"
"The 'optimize' function is the main component of this module.  It sets up\n"
"and runs the TAO optimization.  Two parameters are required:\n\n"
"  x - the initial solution (list)\n"
"  ogfun - objective/gradient function\n\n"
"Other (optional) parameters:\n\n"
"  xub - upper bounds on variables (list)\n"
"  xlb - lower bounds on variables (list)\n"
"  solver - name of TAO solver (string)\n"
"  fatol - absolute convergence tolerance (float)\n"
"  frtol - relative convergence tolerance (float)\n"
"  catol - allowable error in constraints (float)\n"
"  crtol - allowable relative error in constraints (float)\n"
"  verbosity - amount of output (0,1 or 2)\n"
"  maxits - maximum number of iterations (integer)\n\n"
"See source code for default values (\"DEFAULT VALUES\").\n"
"(f,g) = ogfun(x) must calculate the objective and gradient.\n"
"Input:\n\n"
"  x - current solution (list)\n\n"
"Output:\n\n"
"  f - objective value at x (float)\n"
"  g - gradient at x (list)\n\n";

static char getArgs_docs[] =
"Return tuple of all settable parameters (for Optimize).\n"
"Intended mainly for testing/debugging purposes.\n";

static PyMethodDef pytao_funcs[] = {
  {"optimize", (PyCFunction)Optimize, METH_KEYWORDS, optimize_docs},
  {"getArgs", (PyCFunction)GetArgs, METH_KEYWORDS, getArgs_docs},
  {NULL}
};

extern "C" void
initpytao(void)
{
  int numargs = 1;
  char **argvals;
  argvals = (char**) malloc(sizeof(char*)*(numargs+1));
  argvals[0] = "python";
  argvals[1] = NULL;
  TaoInitialize(&numargs,&argvals,NULL,NULL);

  Py_InitModule3("pytao", pytao_funcs, "Python interface to TAO");

  // Might be nice to be able to call this somewhere...
  //TaoFinalize();
}
